(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: relation.ml,v 1.57 2008/02/12 22:13:02 blanqui Exp $ *)

(** {3 Operations on relations of relational types.} *)

open Parsetree
;;

(** {6 Test functions on the type relation.} *)

let is_rewrite r =
  match r.prel_desc with Rewrite _ -> true | _ -> false
and is_commutative r =
  match r.prel_desc with Commutative _ -> true | _ -> false
and is_associative r =
  match r.prel_desc with Associative _ -> true | _ -> false
and is_absorbing r =
  match r.prel_desc with Absorbing _ -> true | _ -> false
and is_neutral r =
  match r.prel_desc with Neutral _ -> true | _ -> false
and is_idempotent r =
  match r.prel_desc with Idempotent _ -> true | _ -> false
and is_inverse r =
  match r.prel_desc with Inverse _ -> true | _ -> false
and is_distributive r =
  match r.prel_desc with Distributive _ -> true | _ -> false
and is_nilpotent r =
  match r.prel_desc with Nilpotent _ -> true | _ -> false
;;

(** {6 Test functions on lists of relations.} *)

let predicate_on_relations p relations =
  match relations.prels_desc with
  | Prels_none | Prels_commented _ -> false
  | Prels_begend rls -> List.exists p rls
;;

let has_commutative = predicate_on_relations is_commutative
and has_associative = predicate_on_relations is_associative
and has_idempotent = predicate_on_relations is_idempotent
and has_absorbing = predicate_on_relations is_absorbing
and has_neutral = predicate_on_relations is_neutral
and has_nilpotent = predicate_on_relations is_nilpotent
;;

(** {6 Access function on lists of relations.} *)

let gather_on_relations gather relations =
  match relations.prels_desc with
  | Prels_none | Prels_commented _ -> []
  | Prels_begend rls -> gather rls
;;

let rec gather_rewrite = function
  | {prel_desc = Rewrite x} :: l -> x :: gather_rewrite l
  | _ :: l -> gather_rewrite l
  | [] -> []
;;

let rec gather_distributive = function
  | {prel_desc = Distributive x} :: l -> x :: gather_distributive l
  | _ :: l -> gather_distributive l
  | [] -> []
;;

let rewrite = gather_on_relations gather_rewrite
and distributive = gather_on_relations gather_distributive
;;

let get_opt_of_relations get relations =
  match relations.prels_desc with
  | Prels_none | Prels_commented _ -> None
  | Prels_begend rls -> get rls
;;

let rec get_commutative = function
  | {prel_desc = Commutative s} :: _ -> s
  | _ :: l -> get_commutative l
  | [] -> None
;;

let rec get_associative = function
  | {prel_desc = Associative s} :: _ -> Some s
  | _ :: l -> get_associative l
  | [] -> None
;;

let rec get_absorbing = function
  | {prel_desc = Absorbing s} :: _ -> Some s
  | _ :: l -> get_absorbing l
  | [] -> None
;;

let rec get_neutral = function
  | {prel_desc = Neutral s} :: _ -> Some s
  | _ :: l -> get_neutral l
  | [] -> None
;;

let rec get_idempotent = function
  | {prel_desc = Idempotent s} :: _ -> Some s
  | _ :: l -> get_idempotent l
  | [] -> None
;;

let rec get_nilpotent = function
  | {prel_desc = Nilpotent (_sd, s)} :: _ -> Some s
  | _ :: l -> get_nilpotent l
  | [] -> None
;;

let rec get_inverse = function
  | {prel_desc = Inverse s} :: _ -> Some s
  | _ :: l -> get_inverse l
  | [] -> None
;;

let commutative = get_opt_of_relations get_commutative
and associative = get_opt_of_relations get_associative
and absorbing = get_opt_of_relations get_absorbing
and neutral = get_opt_of_relations get_neutral
and idempotent = get_opt_of_relations get_idempotent
and nilpotent = get_opt_of_relations get_nilpotent
and inverse = get_opt_of_relations get_inverse
;;

let rec is_neutral_gen gen =
  predicate_on_relations (is_neutral_with_gen gen)

and is_neutral_with_gen g rel =
  match rel.prel_desc with Neutral (_, neutr) -> g.pgen_desc = neutr.pgen_desc | _ -> false


(** {6 Precedence on relations.} *)

(* neutral must be before inverse *)

let precedence rel =
  match rel.prel_desc with
  | Rewrite _ ->               10
  | Associative _ ->           20
  | Commutative _ ->           30
  | Division_by_Absorbent _
  | Neutral _
  | Involutive
  | Absorbing _
  | Idempotent _
  | Structure_item _
  | Status _
  | Precedence _ ->            40
  | Distributive _ ->          50
  | Inverse _ ->               60
  | Nilpotent _ ->             70
;;

(** {6 Comparison function on relations} *)

(** We define a comparison function on relations in the spirit of
    Pervasives.compare. We avoid to take into account locations,
    and use precedences of relations when appropriate. *)

let compare_generator g1 g2 =
  Pervasives.compare g1.pgen_desc g2.pgen_desc
(** We compare the name of generator, not their AST. *)
;;

let compare_side s1 s2 =
  if s1 = Both or s2 = Both
  then 0
  else Pervasives.compare s1 s2
(** Both is equal to every side. *)
;;

(** To compare relations,

    Two non algebraic relations (Rewrite or Structure_items)
    are always considered equal.

    Raises the exception Incomparable when comparing commutativity and
    associativity with themselves.

    - we first test the equality of relations,
    - otherwise, we use the precedence to decide the ordering,
    and in case of same precedence we use Pervasives.compare.

    So, we use the following specification:
    r1 = r2 iff
    - r1 & r2 are the same kind of relation &&
    - r1 & r2 have the same arguments (getting rid of locations).

    Then if not r1 = r2, we compare their relative precedences:
    - if precedence of r1 = precedence of r2 then
    r1 and r2 are in their Pervasives.compare order.
    - else (precedence of r1 <> precedence of r2)
    r1 and r2 are in the order of their precedences.
*)

let compare_relation prec r1 r2 =
  match r1.prel_desc, r2.prel_desc with

  | Structure_item _, Structure_item _
  | Rewrite _, Rewrite _
  | Commutative _, Commutative _
  | Associative _, Associative _
  | Involutive, Involutive -> 0

  | Idempotent s1, Idempotent s2 -> compare_side s1 s2

  | Division_by_Absorbent g1, Division_by_Absorbent g2 ->
      compare_generator g1 g2

  | Nilpotent (s1, g1), Nilpotent (s2, g2)
  | Absorbing (s1, g1), Absorbing (s2, g2)
  | Neutral (s1, g1), Neutral (s2, g2) ->
      begin match compare_generator g1 g2 with
	| 0 -> compare_side s1 s2
	| n -> n
      end

  | Distributive (s1, g1, _h1opt, _b1), Distributive (s2, g2, _h2opt, _b2) ->
      begin match compare_generator g1 g2 with
	| 0 -> compare_side s1 s2
	| n -> n
      end

  | Inverse (s1, g1, _), Inverse (s2, g2, _) ->
      begin match compare_generator g1 g2 with
	| 0 -> compare_side s1 s2
	| n  -> n
      end

  | _, _ ->
    match Pervasives.compare (prec r1) (prec r2) with
    | 0 -> Pervasives.compare r1 r2
    | n -> n
;;
