(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: complete.ml,v 1.60.2.1 2008/02/13 16:00:27 bonichon Exp $ *)

open Relation
open Parsetree
open Genr_base
;;

(*Do we use the order on relations as given in the file *)
let get_user_rel_order, set_user_rel_order =
  let rel_order = ref false in
  (fun () -> !rel_order),
  (fun () -> print_string "User Order On\n";rel_order := true)
;;

(* Builds a function definition from a generator definition. *)

let genr_function_definition gdef =
  let (gname, core_types, relations, loc) = gdef in
  let fd_relations =
    match relations.prels_desc with
    | Prels_none | Prels_commented _ -> relations
    | Prels_begend rels ->
     if get_user_rel_order () then relations else
      { relations with
        prels_desc = Prels_begend
          (List.stable_sort (compare_relation precedence) rels) } in
  { fd_generator = { pgen_desc = Longident.Lident gname; pgen_loc = loc };
    fd_relations = fd_relations;
    fd_arity = List.length core_types;
  }
;;

(* Creates a hash table [generator name -> function definition]
   from a list of generator definitions. *)

let create_table gdefs =
  let ht = Hashtbl.create 17 in
  let enter_def gdef =
    let (gname, _core_types, _rels, _loc) = gdef in
    Hashtbl.add ht gname (genr_function_definition gdef) in
  List.iter enter_def gdefs;
  ht
;;

let add_uniq =
  let ob = Buffer.create 128 in
  fun gname r1 ->
    let print_rule () =
      Printf.bprintf ob "Constructor %s: add relation %a.\n"
	gname Pr_otype.relation r1;
      prerr_string (Buffer.contents ob);
      Buffer.clear ob in
    let print_add l = print_rule (); r1 :: l in
    let rec add = function
      | [] -> print_add []
      | (r2 :: rs) as l ->
          begin
            match compare_relation precedence r1 r2 with
            | n when n > 0 -> r2 :: add rs
            | 0 -> l
            | _ -> print_add l
          end
    in add
(* Add property [r1] in a list of properties [l], only if [r1] is
   not already member of [l].
   The resulting list is sorted wrt the predicate
   [compare_relation].
   We assume that the list argument is already sorted. *)
;;

let rec add_rels gname l1 l2 =
  match l1 with
  | [] -> l2
  | r1 :: m1 -> add_rels gname m1 (add_uniq gname r1 l2)
;;

let add_rels_to_relations gname new_rels relations =
  match relations.prels_desc with
  | Prels_none | Prels_commented _ ->
    { relations with
      prels_desc = Prels_begend (add_rels gname new_rels []) }
  | Prels_begend rels ->
    { relations with
      prels_desc = Prels_begend (add_rels gname new_rels rels) }
;;

(* Function for adding properties to generator [g] in a hash table
   [generator name -> function definition]. *)

let update_props ht g new_props =
  let add_dummy_loc prop =
    { prel_desc = prop; prel_loc = Location.none; } in
  let new_rels = List.map add_dummy_loc new_props in
  let gname = Otype.name_of_generator g in
  let fdef = Hashtbl.find ht gname in
  let fdef =
    { fdef with fd_relations =
      add_rels_to_relations gname new_rels fdef.fd_relations } in
  Hashtbl.replace ht gname fdef
;;

(* Adds in [ht] the properties required by the generator [g] and its
   associated properties, according to the relation argument. *)

let update_by_ap ht fdef =

  let g = fdef.fd_generator in
  let relations = fdef.fd_relations in

  (*FIXME? Division is generated in no example...
  let add_division_by_absorbent e props =
    match absorbent relations with
    | None -> props
    | Some (_side, abs) ->
      (* In some cases, the neutral element is equal
         to the absorbent element,
         hence we have not to generate Division_by_Absorbent. *)
      if compare_generator abs e <> 0
      then Division_by_Absorbent abs :: props
      else props in*)

  let e_opt =
    match neutral relations with
    | Some (_side, e) -> Some e
    | None -> None in

  let rec update = function
    | Inverse (_side, inv, None) when Check.is_local_gen inv ->
      (* A true inverse: we need a neutral element. *)
      let e = match e_opt with
        | Some e -> e
        | None ->
            Check.error g.pgen_loc (Check.No_neutral_element g.pgen_desc) in
      (* Properties of inverse. *)
      let props =
        [Involutive; Distributive (Both, g, Some g, Dist_Inverse)] in
      (* If the operator c is commutative then
           rinv re = re (for right inverse)
           linv le = le (for left inverse)
           inv e = e (for inverse Both). *)
      let props =
        (*FIXME: why? if has_commutative relations
        then*) Distributive (Both, e, Some e, Dist_Direct) :: props
	  (*Inverse_Neutral (e, Some e) :: props*)
        (*else props*) in
      (*let props = add_division_by_absorbent e props in*)
      update_props ht inv props

    | Inverse (side, inv, Some a) when Check.is_local_gen inv ->
      (* If a is the neutral element, then inv is a true inverse,
         hence we just have to call update with the proper argument. *)
      if Some a = e_opt
      then update (Inverse (side, inv, None))
      else
      (* Inv is a fake inverse: we treat it especially. *)
      let props =
        [Involutive; ] in
      let props =
        match e_opt with
        | Some e ->
	    Distributive (Both, e, Some a, Dist_Direct)
	    :: Distributive (Both, e, Some e, Dist_Direct) :: props
        | None -> props in
      (*let props = add_division_by_absorbent a props in*)
      update_props ht inv props

    | Nilpotent (_, a) when Check.is_truly_unary_gen g ->
	update_props ht g [Distributive (Both, a, Some a, Dist_Direct)]

    | Nilpotent _ | Inverse _
    | Division_by_Absorbent _ | Distributive _ | Idempotent _
    | Neutral _ | Absorbing _ | Associative _ | Commutative _
    | Rewrite _ | Involutive | Precedence _ | Structure_item _
    | Status _ -> () in

  function relation -> update relation.prel_desc
;;

(* Adds in [ht] the properties required by the properties of [gdefs]. *)

let update_table ht =
  let update _g fdef =
    let relations = fdef.fd_relations in
    let rels =
      match relations.prels_desc with
      | Prels_none | Prels_commented _ -> []
      | Prels_begend rels -> rels in
    List.iter (update_by_ap ht fdef) rels in
  Hashtbl.iter update ht
;;

(* Gives the list of function definitions corresponding to a list of
   generator definitions, after having added all the required
   properties. *)
let genr_function_definitions gdefs =
  let ht = create_table gdefs in
  update_table ht;
  Hashtbl.fold (fun _ fdef l -> fdef :: l) ht []
;;
