(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: otype.ml,v 1.25 2008/02/11 23:11:03 weis Exp $ *)

(** {3 Functions on OCaml types.} *)

open Asttypes
open Parsetree
open Longident
;;

(** Functions on generators. *)

let name_of_longident = function
  | Lident s
  | Ldot (_, s) -> s
  | Lapply _ -> assert false
;;

let name_of_generator g = name_of_longident g.pgen_desc
;;

let rec qualified_name_of_longident = function
  | Lident s -> s
  | Ldot (lid, s) -> qualified_name_of_longident lid ^ "." ^ s
  | Lapply _ -> assert false
;;

let qualified_name_of_generator g = qualified_name_of_longident g.pgen_desc
;;

(** Functions on core types. *)

let mktyp_noloc desc = { ptyp_desc = desc; ptyp_loc = Location.none; }
;;

let var_type s = mktyp_noloc (Ptyp_var s)
;;

let poly_type tname params =
  mktyp_noloc (Ptyp_constr (Lident tname, List.map var_type params))
;;

let constant_type tname = poly_type tname []
;;

let bool_type = constant_type "bool"
;;

let arrow_type ct1 ct2 = mktyp_noloc (Ptyp_arrow ("", ct1, ct2))
;;

let tuple_type cts = mktyp_noloc (Ptyp_tuple cts)
;;

let list_type ct = mktyp_noloc (Ptyp_constr (Lident "list", [ct]))
;;

let type_of_eq tname params =
  let ct = poly_type tname params in
  arrow_type ct (arrow_type ct bool_type)
;;

let projection_type tname params ctrep =
  let ct = poly_type tname params in
  arrow_type ct ctrep
;;

let injection_type tname params ctrep =
  let ct = poly_type tname params in
  arrow_type ctrep ct
;;

let is_list_type ct =
  match ct.ptyp_desc with
  | Ptyp_constr (Lident "list", [_ct]) (* list may have been redefined ? *)
  | Ptyp_constr (Ldot (Lident "Pervasives", "list"), [_ct]) -> true
  | _ -> false
;;

let list_argument ct = 
  match ct.ptyp_desc with
  | Ptyp_constr (Lident "list", [ct]) (* list may have been redefined ? *)
  | Ptyp_constr (Ldot (Lident "Pervasives", "list"), [ct]) -> ct
  | _ -> raise (Invalid_argument "Otype.list_argument")
;;

(** Functions on relations. *)

let comment_relations_in_type_kind = function
  | Ptype_variant cstrs ->
      let comment_rels (gid, tycl, rels, loc) =
        let new_rels =
          { rels with prels_desc = Prels_commented rels } in
        (gid, tycl, new_rels, loc) in
      Ptype_variant (List.map comment_rels cstrs)
  | Ptype_record flds ->
      let comment_rels (fid, mflag, fty, rels, loc) =
        let new_rels =
          { rels with prels_desc = Prels_commented rels } in
        (fid, mflag, fty, new_rels, loc) in
      Ptype_record (List.map comment_rels flds)
  | Ptype_abstract rels ->
      let new_rels =
        { rels with prels_desc = Prels_commented rels } in
      Ptype_abstract new_rels
;;

let comment_relations_in_type_declaration td =
  { td with ptype_kind = comment_relations_in_type_kind td.ptype_kind; }
;;

(** Functions on type declarations. *)

let make_public_type_declaration td =
  { td with ptype_private = Public }
;;

let get_structure_items_of_type_declaration td =

  let rec get_structure_items_of_rls accu = function
    | [] -> accu
    | { prel_desc = Structure_item struct_item } :: rls ->
      get_structure_items_of_rls (struct_item :: accu) rls
    | { prel_desc = _ } :: rls ->
      get_structure_items_of_rls accu rls in

  let rec get_structure_items_of_relations accu rels =
    match rels.prels_desc with
    | Prels_none -> accu
    | Prels_commented rels -> get_structure_items_of_relations accu rels
    | Prels_begend rls -> get_structure_items_of_rls accu rls in

  let get_structure_items_of_generators accu cstrs =
    List.fold_left
      (fun str_items (_gid, _ctys, rels, _loc) ->
         get_structure_items_of_relations [] rels @ str_items)
      accu cstrs in

  let get_structure_items_of_fields accu flds =
    List.fold_left
      (fun str_items (_fid, _mflag, _fty, rels, _loc) ->
         get_structure_items_of_relations [] rels @ str_items)
      accu flds in

  let get_structure_items_of_type_kind = function
    | Ptype_abstract relations ->
      get_structure_items_of_relations [] relations
    | Ptype_variant cstrs ->
      get_structure_items_of_generators [] cstrs
    | Ptype_record flds ->
      get_structure_items_of_fields [] flds in

  List.rev (get_structure_items_of_type_kind td.ptype_kind)
;;
